#include <bout/bout_types.hxx>
#include <bout/field3d.hxx>

void c_get_f3d_all(const Field3D * f, double * data);
void c_set_all(Field3D * f, const Field3D * rhs);
void c_set_all(Field3D * f, const double * data);
void c_set_all(Field3D * f, double data);
void c_set_part(Field3D* field, const double* data, const size_t* indices, size_t num);
void c_get_part(Field3D* field, double* data, const size_t* indices, size_t num);


Field3D * fadd( Field3D*,Field3D*);
Field3D * fadd( Field3D*,double);
Field3D * fadd( double,Field3D*);
void fiadd( Field3D*,Field3D*);
void fiadd( Field3D*,double);
Field3D * fadd( Field3D*,Field2D*);
Field3D * fadd( Field2D*,Field3D*);
void fiadd( Field3D*,Field2D*);
Field3D * fmul( Field3D*,Field3D*);
Field3D * fmul( Field3D*,double);
Field3D * fmul( double,Field3D*);
void fimul( Field3D*,Field3D*);
void fimul( Field3D*,double);
Field3D * fmul( Field3D*,Field2D*);
Field3D * fmul( Field2D*,Field3D*);
void fimul( Field3D*,Field2D*);
Field3D * ftruediv( Field3D*,Field3D*);
Field3D * ftruediv( Field3D*,double);
Field3D * ftruediv( double,Field3D*);
void fitruediv( Field3D*,Field3D*);
void fitruediv( Field3D*,double);
Field3D * ftruediv( Field3D*,Field2D*);
Field3D * ftruediv( Field2D*,Field3D*);
void fitruediv( Field3D*,Field2D*);
Field3D * fdiv( Field3D*,Field3D*);
Field3D * fdiv( Field3D*,double);
Field3D * fdiv( double,Field3D*);
void fidiv( Field3D*,Field3D*);
void fidiv( Field3D*,double);
Field3D * fdiv( Field3D*,Field2D*);
Field3D * fdiv( Field2D*,Field3D*);
void fidiv( Field3D*,Field2D*);
Field3D * fsub( Field3D*,Field3D*);
Field3D * fsub( Field3D*,double);
Field3D * fsub( double,Field3D*);
void fisub( Field3D*,Field3D*);
void fisub( Field3D*,double);
Field3D * fsub( Field3D*,Field2D*);
Field3D * fsub( Field2D*,Field3D*);
void fisub( Field3D*,Field2D*);
Field3D c_minus(Field3D a){return -a;};

void c_get_f3d_part(const Field3D * f, double * data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy, size_t zs, size_t ze, int dz);
void c_set_f3d_part(Field3D * f, const double * data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy, size_t zs, size_t ze, int dz);
void c_set_f3d_part_(Field3D * f, const double data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy, size_t zs, size_t ze, int dz);
void c_get_f2d_all(const Field2D * f, double * data);
void c_set_all(Field2D * f, const Field2D * rhs);
void c_set_all(Field2D * f, const double * data);
void c_set_all(Field2D * f, double data);
void c_set_part(Field2D* field, const double* data, const size_t* indices, size_t num);
void c_get_part(Field2D* field, double* data, const size_t* indices, size_t num);


Field2D * fadd( Field2D*,Field2D*);
Field2D * fadd( Field2D*,double);
Field2D * fadd( double,Field2D*);
void fiadd( Field2D*,Field2D*);
void fiadd( Field2D*,double);
Field2D * fmul( Field2D*,Field2D*);
Field2D * fmul( Field2D*,double);
Field2D * fmul( double,Field2D*);
void fimul( Field2D*,Field2D*);
void fimul( Field2D*,double);
Field2D * ftruediv( Field2D*,Field2D*);
Field2D * ftruediv( Field2D*,double);
Field2D * ftruediv( double,Field2D*);
void fitruediv( Field2D*,Field2D*);
void fitruediv( Field2D*,double);
Field2D * fdiv( Field2D*,Field2D*);
Field2D * fdiv( Field2D*,double);
Field2D * fdiv( double,Field2D*);
void fidiv( Field2D*,Field2D*);
void fidiv( Field2D*,double);
Field2D * fsub( Field2D*,Field2D*);
Field2D * fsub( Field2D*,double);
Field2D * fsub( double,Field2D*);
void fisub( Field2D*,Field2D*);
void fisub( Field2D*,double);
Field2D c_minus(Field2D a){return -a;};

void c_get_f2d_part(const Field2D * f, double * data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy);
void c_set_f2d_part(Field2D * f, const double * data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy);
void c_set_f2d_part_(Field2D * f, const double data, size_t xs, size_t xe, int dx, size_t ys, size_t ye, int dy);
Field3D c_mul_Vector3D_Vector3D(const Vector3D * a, const Vector3D * b){ return (*a) * (*b) ;};
Field2D c_mul_Vector2D_Vector2D(const Vector2D * a, const Vector2D * b){ return (*a) * (*b) ;};

Mesh * c_get_global_mesh();

Field3D c_Grad_perp_dot_Grad_perp(Field3D a, Field3D b){
  return Grad_perp(a)*Grad_perp(b);
}

void c_mesh_normalise(Mesh * , double);


class PythonModelCallback{
 public:
  typedef void (*Method)(void *param, void *user_data);
  PythonModelCallback(Method method, void *user_data){
    _method=method;
    _user_data=user_data;
  }
  virtual ~PythonModelCallback(){};
  void cy_execute(void *parameter){
    _method(parameter, _user_data);
  }
 private:
  Method _method; // cython function
  void *_user_data; // actual function
};

class PythonModel: public PhysicsModel{
protected:
  int init(bool restarting) override{
    long aligned_restarting = restarting;
    if (_init){
      _init->cy_execute(&aligned_restarting);
    }
    return 0;
  };
  int rhs(BoutReal t) override {
    _rhs->cy_execute( &t);
    return 0;
  };
public:
  PythonModel():_init(nullptr), _rhs(nullptr){};
  void pyinit(){
    solver = Solver::create();
    solver->setModel(this);
    bout_monitor = new BoutMonitor();
    solver->addMonitor(bout_monitor, Solver::BACK);
  };
  void free(){
    delete bout_monitor;
  };
  void set_init_func(PythonModelCallback * __init){
    _init=__init;
  };
  void set_rhs_func(PythonModelCallback * __rhs){
    _rhs=__rhs;
  };
  void solve(){
    solver->solve();
  }
  Solver * getSolver(){
    return solver.get();
  };
private:
  PythonModelCallback * _init;
  PythonModelCallback * _rhs;
  std::unique_ptr<Solver> solver{nullptr};
  Monitor * bout_monitor;
};

void throw_BoutException(std::string err){
  throw BoutException(err);
}
